import numpy as np
import torch


class FWD_J1(torch.autograd.Function):
    """Differentiable function doing forward DTCWT at level 1.
    Returns low-pass (-pi/4 to pi/4) and high-pass (pi/4 to 3pi/4) as a pair.
    """

    @staticmethod
    def forward(ctx, x, h0, h1, skip_hps):
        """
        Forward dual-tree complex wavelet transform at level 1.

        Args:
            ctx is the DTCWTForward object
            x is the input Tensor
            h0 is the low-pass analysis filter
            h1 is the high-pass analysis filter
            skip_hps: if True, skip high-pass filtering

        Returns:
            lo: low-pass output (-pi/4 to pi/4)
            hi: high-pass output (pi/4 to 3pi/4)
        """

        # Replicate filters along the channel dimension
        b, ch, T = x.shape
        h0_rep = h0.repeat(ch, 1, 1)
        h1_rep = h1.repeat(ch, 1, 1)
        ctx.save_for_backward(h0_rep, h1_rep)

        # Apply low-pass filtering
        h0_padding = h0.shape[-1] // 2
        lo = torch.nn.functional.conv1d(
            x, h0_rep, padding=h0_padding, groups=ch)

        # Apply high-pass filtering. If skipped, create an empty array.
        if skip_hps:
            hi = x.new_zeros([])
        else:
            h1_padding = h1.shape[-1] // 2
            hi = torch.nn.functional.conv1d(
                x, h1_rep, padding=h1_padding, groups=ch)

        # Return low-pass (x_phi) and high-pass (x_psi) pair
        return lo, hi


class FWD_J2PLUS(torch.autograd.Function):
    """
    Differentiable function doing forward DTCWT at scales j>1.

    The real part of bp corresponds to the high-pass output of
    tree a whereas the imaginary part of bp corresponds to the
    high-pass output of tree b."""

    @staticmethod
    def forward(ctx, x_a, x_b, h0a, h1a, h0b, h1b, skip_hps):
        """
        Forward dual-tree complex wavelet transform at levels 2 and coarser.

        Args:
            ctx: DTCWTForward object
            x_a: input Tensor of tree a
            x_b: input Tensor of tree b
            h0a: low-pass filter of tree a (real part)
            h1a: high-pass filter of tree a (real part)
            h0b: low-pass filter of tree b (imaginary part)
            h1b: high-pass filter of tree b (imaginary part)
            skip_hps: if True, skip high-pass filtering

        Returns:
            lo_a: low-pass output from tree a
            lo_b: low-pass output from tree b
            bp: band-pass output from both trees."""

        # Replicate filters along the channel dimension
        b, ch, T = x_a.shape
        h0a_rep = h0a.repeat(ch, 1, 1)
        h1a_rep = h1a.repeat(ch, 1, 1)
        h0b_rep = h0b.repeat(ch, 1, 1)
        h1b_rep = h1b.repeat(ch, 1, 1)
        ctx.save_for_backward(h0a_rep, h1a_rep, h0b_rep, h1b_rep)

        # Apply low-pass filtering on trees a (real) and b (imaginary).
        h0a_padding = h0a.shape[-1] // 2
        h0b_padding = h0b.shape[-1] // 2
        lo_a = torch.nn.functional.conv1d(
            x_a, h0a_rep, padding=h0a_padding, groups=ch)
        lo_b = torch.nn.functional.conv1d(
            x_b, h0b_rep, padding=h0b_padding, groups=ch)

        # Apply high-pass filtering. If skipped, create an empty array.
        if skip_hps:
            bp = x_a.new_zeros([])
        else:
            h1a_padding = h1a.shape[-1] // 2
            h1b_padding = h1b.shape[-1] // 2
            bp_a = torch.nn.functional.conv1d(
                lo_a, h1a_rep, padding=h1a_padding, groups=ch
            )
            bp_b = torch.nn.functional.conv1d(
                lo_b, h1b_rep, padding=h1b_padding, groups=ch
            )
            bp = np.sqrt(1/2) * (bp_a + 1j * bp_b)

        # Subsample
        lo_a = np.sqrt(1/2) * lo_a[:, :, ::2]
        lo_b = np.sqrt(1/2) * lo_b[:, :, ::2]

        # Return low-pass output from both trees separately, and band-pass
        # output in conjunction:
        # real part for tree a and imaginary part for tree b.
        return lo_a, lo_b, bp


class INV_J1(torch.autograd.Function):
    """Differentiable function doing inverse DTCWT at level 1.
    Returns a full-band 1-D signal from a low-pass and a high-pass component.
    """

    @staticmethod
    def forward(ctx, lo, hi, g0, g1):
        """
        Inverse dual-tree complex wavelet transform at level 1.

        Args:
            ctx is the DTCWTInverse object
            lo is the low-pass input (-pi/4 to pi/4)
            hi is the high-pass input (pi/4 to 3pi/4)
            g0 is the low-pass synthesis filter
            g1 is the high-pass synthesis filter

        Returns:
            x: reconstructed 1-D signal (possibly multichannel)
        """
        # Replicate filters along the channel dimension
        b, ch, T = lo.shape
        g0_rep = g0.repeat(ch, 1, 1)
        g1_rep = g1.repeat(ch, 1, 1)
        ctx.save_for_backward(g0_rep, g1_rep)

        # Apply dual low-pass filtering
        g0_padding = g0.shape[-1] // 2
        x0 = torch.nn.functional.conv1d(lo, g0_rep, padding=g0_padding)

        # Apply dual high-pass filtering
        g1_padding = g1.shape[-1] // 2
        x1 = torch.nn.functional.conv1d(hi, g1_rep, padding=g1_padding)

        # Mix low-pass and high-pass contributions
        x = x0 + x1
        return x


class INV_J2PLUS(torch.autograd.Function):
    """Differentiable function doing inverse DTCWT at levels >1.
    Returns a broadband 1-D signal from a low-pass and a high-pass component.
    """

    @staticmethod
    def forward(ctx, lo_a, lo_b, bp, g0a, g1a, g0b, g1b):
        """
        Inverse dual-tree complex wavelet transform at levels 2 and coarser.

        Args:
            ctx is the DTCWTInverse object
            lo_a is the low-pass output from tree a (real part)
            lo_b is the low-pass output from tree b (imaginary part)
            bp is the band-pass output (complex-valued)
            g0a is the dual low-pass filter of tree a (real part)
            g1a is the dual high-pass filter of tree a (real part)
            g0b is the dual low-pass filter of tree b (imaginary part)
            g1b is the dual high-pass filter of tree b (imaginary part)

        Returns:
            x_a: reconstructed 1-D signal from tree a (real part)
            x_b: reconstructed 1-D signal from tree a (imaginary part)
        """
        # Replicate filters along the channel dimension
        b, ch, T = lo_a.shape
        g0a_rep = g0a.repeat(ch, 1, 1)
        g1a_rep = g1a.repeat(ch, 1, 1)
        g0b_rep = g0b.repeat(ch, 1, 1)
        g1b_rep = g1b.repeat(ch, 1, 1)
        ctx.save_for_backward(g0a_rep, g1a_rep, g0b_rep, g1b_rep)

        # Apply dual low-pass filtering on trees a (real) and b (imaginary)
        g0a_padding = g0a.shape[-1] // 2
        g0b_padding = g0b.shape[-1] // 2
        x0a = torch.nn.functional.conv1d(lo_a, g0a_rep, padding=g0a_padding)
        x0b = torch.nn.functional.conv1d(lo_b, g0b_rep, padding=g0a_padding)

        # Upsample by inserting zeros every other sample
        x0a_zeros = torch.zeros_like(x0a)
        x0a_up = torch.stack((x0a, x0a_zeros), -1).reshape(b, ch, 2 * T)
        x0b_zeros = torch.zeros_like(x0b)
        x0b_up = torch.stack((x0b, x0b_zeros), -1).reshape(b, ch, 2 * T)

        # Apply dual high-pass filtering on band-pass input (complex-valued)
        g1a_padding = g1a.shape[-1] // 2
        g1b_padding = g1b.shape[-1] // 2
        bp_a = torch.real(bp)
        bp_b = torch.imag(bp)
        x1a = torch.nn.functional.conv1d(bp_a, g1a_rep, padding=g1a_padding)
        x1b = torch.nn.functional.conv1d(bp_b, g1b_rep, padding=g1b_padding)

        # Mix low-pass and high-pass contributions
        x_a = x0a + x1a
        x_b = x0b + x1b
        return x_a, x_b



import numpy as np
import pytorch_wavelets as pytw
import torch.nn

from .lowlevel import prep_filt
from .transform_funcs import FWD_J1, FWD_J2PLUS


class DTCWTForward(torch.nn.Module):
    """Performs a DTCWT forward decomposition of a PyTorch tensor containing
    1-D signals, potentially multichannel.

    Args:
        level1 (str): One of 'antonini', 'legall', 'near_sym_a', 'near_sym_b'.
            Specifies the first-level biorthogonal wavelet filters.
        qshift (str): One of 'qshift_06', 'qshift_a', 'qshift_b', 'qshift_c',
            'qshift_d'.  Specifies the second-level quarter-shift filters.
        J (int): Number of levels (octaves) of decomposition. Default is 8.
        skip_hps (bools): List of bools of length J which specify whether or
            not to calculate the bandpass outputs at the given scale.
            skip_hps[0] is for the first scale. Can be a single bool in which
            case that is applied to all scales.
        include_scale (bool): If true, return the bandpass outputs. Can also be
            a list of length J specifying which lowpasses to return. I.e. if
            [False, True, True], the forward call will return the second and
            third lowpass outputs, but discard the lowpass from the first level
            transform.
        alternate_gh (bool): If True (default), alternates between filter pairs
            (h0, h1) and (g0, g1) depending on odd vs. even wavelet scale j.
            Otherwise, uses (h0, h1) only. See Selesnick et al. 2005 for details.
    """

    def __init__(
        self,
        level1="near_sym_a",
        qshift="qshift_a",
        J=8,
        skip_hps=False,
        include_scale=False,
        alternate_gh=True
    ):
        # Instantiate PyTorch NN Module
        super().__init__()

        # Store metadata
        self.level1 = level1
        self.qshift = qshift
        self.J = J
        self.alternate_gh = alternate_gh

        # Load first-level biorthogonal wavelet filters from disk.
        # h0o is the low-pass filter.
        # h1o is the high-pass filter.
        h0o, _, h1o, _ = pytw.dtcwt.coeffs._load_from_file(
            level1, ("h0o", "g0o", "h1o", "g1o")
        )
        self.register_buffer("h0o", prep_filt(h0o))
        self.register_buffer("h1o", prep_filt(h1o))

        # Load higher-level quarter-shift wavelet filters from disk.
        # h0a is the low-pass filter from tree a (real part).
        # h0b is the low-pass filter from tree b (imaginary part).
        # g0a is the low-pass dual filter from tree a (real part).
        # g0b is the low-pass dual filter from tree b (imaginary part).
        # h1a is the high-pass filter from tree a (real part).
        # h1b is the high-pass filter from tree b (imaginary part).
        # g1a is the high-pass dual filter from tree a (real part).
        # g1b is the high-pass dual filter from tree b (imaginary part).
        h0a, h0b, g0a, g0b, h1a, h1b, g1a, g1b = pytw.dtcwt.coeffs._load_from_file(
            qshift, ("h0a", "h0b", "g0a", "g0b", "h1a", "h1b", "g1a", "g1b")
        )
        self.register_buffer("h0a", prep_filt(h0a))
        self.register_buffer("h0b", prep_filt(h0b))
        self.register_buffer("g0a", prep_filt(g0a))
        self.register_buffer("g0b", prep_filt(g0b))
        self.register_buffer("h1a", prep_filt(h1a))
        self.register_buffer("h1b", prep_filt(h1b))
        self.register_buffer("g1a", prep_filt(g1a))
        self.register_buffer("g1b", prep_filt(g1b))

        # Parse the "skip_hps" argument for skipping finest scales.
        if isinstance(skip_hps, (list, tuple, np.ndarray)):
            self.skip_hps = skip_hps
        else:
            self.skip_hps = [skip_hps,] * self.J

        # Parse the "include_scale" argument for including other low-pass
        # outputs in addition to the coarsest scale.
        if isinstance(include_scale, (list, tuple, np.ndarray)):
            self.include_scale = include_scale
        else:
            self.include_scale = [include_scale,] * self.J

    def forward(self, x):
        """Forward Dual-Tree Complex Wavelet Transform (DTCWT) of a 1-D signal.

        Args:
            x (PyTorch tensor): Input data. Should be a tensor of shape
                `(B, C, T)` where B is the batch size, C is the number of
                channels and T is the number of time samples.
                Note that T must be a multiple of 2**J, where J is the number
                of wavelet scales (see documentation of DTCWTForward constructor).

        Returns:
            (yl, yh): tuple of low-pass (yl) and band-pass (yh) coefficients.
                If include_scale is True (see DTCWTForward constructor), yl is a
                list of low-pass coefficients at all wavelet scales 1 to (J-1).
                Otherwise (default), yl is a real-valued PyTorch tensor of shape
                `(B, C, T/2**(J-1))`.
                Conversely, yh is a list of PyTorch tensors with J elements,
                containing the band-pass coefficients at all wavelets scales 1 to
                (J-1). These tensors are complex-valued and have shapes:
                `(B, C, T)`, `(B, C, T/2)`, `(B, C, T/4)`, etc."""

        # Initialize lists of empty arrays with same dtype as input
        x_phis = [x.new_zeros([]),] * self.J
        x_psis = [x.new_zeros([]),] * self.J

        # Assert that the length of x is a multiple of 2**J
        T = x.shape[-1]
        assert T % (2 ** self.J) == 0

        ## LEVEL 1 ##
        x_phi, x_psi = FWD_J1.apply(x, self.h0o, self.h1o, self.skip_hps[0])
        x_psis[0] = x_psi
        if self.include_scale[0]:
            x_phis[0] = x_phi

        ## LEVEL 2 AND GREATER ##
        # Apply multiresolution pyramid by looping over j from fine to coarse
        x_phia = x_phi
        x_phib = x_phi
        for j in range(1, self.J):
            if (j%2 == 1) and self.alternate_gh:
                # Pick the dual filters g0a, g1a, etc. instead of h0a, h1a, etc.
                h0a, h1a, h0b, h1b = self.g0a, self.g1a, self.g0b, self.g1b
            else:
                h0a, h1a, h0b, h1b = self.h0a, self.h1a, self.h0b, self.h1b

            x_phia, x_phib, x_psi = FWD_J2PLUS.apply(
                x_phia, x_phib, h0a, h1a, h0b, h1b, self.skip_hps[j]
            )

            if (j%2 == 1) and self.alternate_gh:
                # The result is anti-analytic in the Hilbert sense.
                # We conjugate the result to bring the spectrum back to (0, pi).
                # This is purely by convention and for consistency through j.
                x_psi = torch.conj(x_psi)

            x_psis[j] = x_psi
            if self.include_scale[j]:
                x_phis[j] = x_phia + 1j * x_phib

        # If at least one of the booleans in the list include_scale is True,
        # return the list x_phis as yl. Otherwise, return the last x_phi.
        if True in self.include_scale:
            return x_phis, x_psis
        else:
            return (x_phia + 1j * x_phib), x_psis



import numpy as np
import torch

import murenn.dtcwt.lowlevel


def test_prep_filt():
    x = np.array([1, 2, 3])[:, np.newaxis]
    y = murenn.dtcwt.lowlevel.prep_filt(x)
    assert isinstance(y, torch.Tensor)
    assert list(y.shape) == [1, 1, len(x)]



import pytest
import torch

import murenn


@pytest.mark.parametrize("skip_hps", [False, [0, 1]])
@pytest.mark.parametrize("include_scale", [False, [7, 8]])
def test_dtcwt(skip_hps, include_scale):
    W = murenn.DTCWT(skip_hps=skip_hps, include_scale=include_scale)
    assert isinstance(W, torch.nn.Module)



import torch


def prep_filt(h):
    """Convert a 1-D filter from pytorch_wavelets (NumPy row vector)
    into a PyTorch object (1x1xT Tensor) for MuReNN."""
    # Note that we drop the trailing singleton dimension because
    # this filter will be used on 1-D signals, not 2-D images.
    # Conversely, we prepend two singleton dimensions: one for
    # the signal index (in the batch) and the other for the
    # channel index (for multimodal time series).
    return torch.tensor(h[None, None, :, 0], dtype=torch.float32)



"""
MuReNN: MultiResolution Neural Networks
=======================================

murenn is a Python module which integrates efficient operators
for multiresolution analysis into a differentiable computing
framework.
"""

# List of top-level public names.
__all__ = ["DTCWT", "DTCWTForward"]


# Submodule imports
from .dtcwt.transform1d import DTCWTForward
from .version import version as __version__

# PytW-like aliases
DTCWT = DTCWTForward



from importlib.machinery import SourceFileLoader
from setuptools import setup


version = SourceFileLoader('murenn.version', 'murenn/version.py').load_module()

with open('README.md', 'r') as fdesc:
    long_description = fdesc.read()

setup(
    name='murenn',
    version=version.version,
    description='Multiresolution neural networks',
    author='The Kymatio consortium',
    url='https://gitlab.univ-nantes.fr/kymatio/murenn',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'License :: OSI Approved :: ISC License (ISCL)',
        'Programming Language :: Python',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Multimedia :: Sound/Audio :: Analysis',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    keywords='wavelets, deep learning, signal processing, audio, speech, music',
    license='3-clause BSD',
    install_requires=['pytorch_wavelets', 'PyWavelets'],
    dependency_links=[
        'pytorch_wavelets @ git+ssh://git@github.com/fbcotter/pytorch_wavelets@1.1.0#egg=pytorch_wavelets-1.1.0'
    ],
)
